# -*- coding: UTF-8 -*-
import requests
from pprint import pprint as pp
import json
import re
import sys
import datetime
import logging
import logging.config
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class LookerApi(object):
    def __init__(self, id, secret, host, port, access_token):
        self.api_logger = logging.getLogger('lookerapi')
        self.id = id
        self.secret = secret
        self.host = host
        self.port = port
        self.access_token = access_token

        self.session = requests.Session()
        self.session.verify = False

        self.session.headers.update({'Authorization': 'token %s' %
                                    access_token})

        # if not valid anymore, authenticate again
        if self.__get_me() == 401:
            self.api_logger.warning('Existing auth token has expired')
            self.auth()

    def get_access_token(self):
        return self.access_token

    def auth(self):
        self.api_logger.info('Authenticating')
        url = 'https://{}:{}/api/3.0/{}'.format(self.host, self.port, 'login')
        params = {'client_id': self.id,
                  'client_secret': self.secret
                  }
        self.api_logger.info('Request to %s => POST /api/3.0/login, %s',
                             self.host, {'client_id': params['client_id'],
                                         'client_secret': "[FILTERED]"})
        r = self.session.post(url, params=params, timeout=60)
        access_token = r.json().get('access_token')
        self.session.headers.update({'Authorization': 'token '
                                    '{}'.format(access_token)})
        if r.status_code == requests.codes.ok:
            self.api_logger.info('Request Complete: %s', r.status_code)
            self.access_token = access_token
        else:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print('Authentication Error: Check supplied credentials.')
            sys.exit(1)

        return

# GET /user - meant for use by the class itself
    def __get_me(self):
        self.api_logger.info('Trying to auth in using existing auth token')
        url = 'https://{}:{}/api/3.0/user'.format(self.host, self.port)
        self.api_logger.info('Request to %s => POST /api/3.0/user', self.host)
        try:
            r = self.session.get(url, timeout=1)
        except Exception as e:
            self.api_logger.error(e)
            print('Connection timed out. Please confirm the hostname')
            sys.exit(1)
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.status_code

# GET /lookml_models/
    def get_models(self, fields={}):
        url = 'https://{}:{}/api/3.0/{}'.format(self.host,
                                                self.port,
                                                'lookml_models')
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/lookml_models, %s',
                             self.host,
                             params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.error('Request Complete: %s', r.status_code)
            raise(e)
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /lookml_models/{{NAME}}
    def get_model(self, model_name=None, fields={}):
        url = 'https://{}:{}/api/3.0/{}/{}'.format(self.host,
                                                   self.port,
                                                   'lookml_models',
                                                   model_name)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/lookml_models/%s,'
                             ' %s', self.host, model_name, params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.error('Request Complete: %s', r.status_code)
            raise(e)
        self.api_logger.info('Request Complete: %s', r.status_code)
        return [r.json()]

# GET /lookml_models/{{NAME}}/explores/{{NAME}}
    def get_explore(self, model_name=None, explore_name=None, fields={}):
        url = 'https://{}:{}/api/3.0/{}/{}/{}/{}'.format(self.host,
                                                         self.port,
                                                         'lookml_models',
                                                         model_name,
                                                         'explores',
                                                         explore_name)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/lookml_models/%s'
                             '/explores/%s, %s', self.host, model_name,
                             explore_name, params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.error('Request Complete: %s', r.status_code)
            return []
        self.api_logger.info('Request Complete: %s', r.status_code)
        return [r.json()]

# GET /projects
    def get_projects(self, fields={}):
        url = 'https://{}:{}/api/3.0/{}'.format(self.host,
                                                self.port,
                                                'projects')
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/projects, %s',
                             self.host,
                             params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            raise(e)
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /projects/{project_id}
    def get_project(self, project_id=None, fields={}):
        url = 'https://{}:{}/api/3.0/{}/{}'.format(self.host,
                                                   self.port,
                                                   'projects',
                                                   project_id)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/projects/%s, %s',
                             self.host,
                             project_id,
                             params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.error('Request Complete: %s', r.status_code)
            raise e
        self.api_logger.info('Request Complete: %s', r.status_code)
        return [r.json()]

# GET /projects/{project_id}/files
    def get_project_files(self, project=None, fields={}):
        url = 'https://{}:{}/api/3.0/{}/{}/{}'.format(self.host,
                                                      self.port,
                                                      'projects',
                                                      project,
                                                      'files')
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/projects/%s/files,'
                             ' %s', self.host, project, params)
        r = self.session.get(url, params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print('Project not found: %s', project)
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# POST /queries/run/{result_format}
    def run_inline_query(self, result_format, body, fields={}):
        url = 'https://{}:{}/api/3.0/{}/{}/{}'.format(self.host,
                                                      self.port,
                                                      'queries',
                                                      'run',
                                                      result_format)
        params = fields
        self.api_logger.info('Request to %s => POST /api/3.0/queries/run/%s, '
                             '%s', self.host, result_format, params)
        self.api_logger.info('Query params=%s', body)
        r = self.session.post(url, json.dumps(body), params=params, timeout=60)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# PATCH session
    def update_session(self, mode):
        url = 'https://{}:{}/api/3.0/{}'.format(self.host,
                                                self.port,
                                                'session')
        body = {'workspace_id': str(mode)}
        self.api_logger.info('Request to %s => PATCH /api/3.0/session, %s',
                             self.host,
                             body)
        r = self.session.patch(url, json=body)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET session
    def get_session(self, fields={}):
        url = 'https://{}:{}/api/3.0/{}'.format(self.host,
                                                self.port,
                                                'session')
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/session, %s',
                             self.host,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /projects/{project_id}/git_connection_tests
    def git_connection_tests(self, project_id, fields={}):
        url = ('https://{}:{}/api/3.0/projects/{}/'
               'git_connection_tests').format(self.host, self.port, project_id)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/projects/%s/'
                             'git_connection_tests, %s',
                             self.host,
                             project_id,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /projects/{project_id}/git_connection_tests/{test_id}
    def run_git_connection_test(self, project_id, test_id, fields={}):
        url = ('https://{}:{}/api/3.0/projects/{}/git_connection_tests/'
               '{}').format(self.host, self.port, project_id, test_id)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/projects/%s/'
                             'git_connection_tests/%s, %s',
                             self.host,
                             project_id,
                             test_id,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /connections
    def get_connections(self, fields={}):
        url = 'https://{}:{}/api/3.0/connections'.format(self.host, self.port)
        params = fields
        self.api_logger.info('Request to %s => GET /api/3.0/connections, %s',
                             self.host,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# PUT /connections/{connection_name}/test
    def test_connection(self, connection, fields={}):
        url = 'https://{}:{}/api/3.0/connections/{}/test'.format(self.host,
                                                                 self.port,
                                                                 connection)
        params = fields
        self.api_logger.info('Request to %s => POST /api/3.0/connections/'
                             '%s/test, %s', self.host, connection, params)
        r = self.session.put(url, params=params)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /legacy_features
    def get_legacy_features(self, fields={}):
        url = 'https://{}:{}/api/3.0/legacy_features'.format(self.host,
                                                             self.port)
        params = fields
        self.api_logger.info('Request to %s => POST /api/3.0/legacy_features,'
                             ' %s', self.host, params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /integrations
    def get_integrations(self, fields={}):
        url = 'https://{}:{}/api/3.0/integrations'.format(self.host, self.port)
        params = fields
        self.api_logger.info('Request to %s => POST /api/3.0/integrations, %s',
                             self.host,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()

# GET /versions
    def get_version(self, fields={}):
        url = 'https://{}:{}/api/3.0/versions'.format(self.host, self.port)
        params = fields
        self.api_logger.info('Request to %s => POST /api/3.0/versions, %s',
                             self.host,
                             params)
        r = self.session.get(url)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.api_logger.warning('Request Complete: %s', r.status_code)
            print("Error: " + str(e))
            return
        self.api_logger.info('Request Complete: %s', r.status_code)
        return r.json()



import threading
import sys
import time


class SpinnerThread(threading.Thread):

    def __init__(self):
        super().__init__(target=self._spin)
        self._stopevent = threading.Event()

    def stop(self):
        sys.stdout.write('\b')
        self._stopevent.set()

    def _spin(self):

        while not self._stopevent.isSet():
            for t in '|/-\\':
                sys.stdout.write(t)
                sys.stdout.flush()
                time.sleep(0.1)
                sys.stdout.write('\b')


class Spinner(object):
    def __enter__(self):
        self.spinner = SpinnerThread()
        self.spinner.start()

    def __exit__(self, exc_type, exc_value, tb):
        self.spinner.stop()



import logging
from henry.modules.fetcher import Fetcher as fetcher
from henry.modules import styler
from tabulate import tabulate
import json


class Analyze(fetcher):
    def __init__(self, looker):
        super().__init__(looker)
        self.analyze_logger = logging.getLogger('analyze')

    def analyze(self, **kwargs):
        format = 'plain' if kwargs['plain'] else 'psql'
        headers = '' if kwargs['plain'] else 'keys'
        p = kwargs['project'] if 'project' in kwargs.keys() else None
        m = kwargs['model'] if 'model' in kwargs.keys() else None
        self.analyze_logger.info('Analyzing %s', kwargs['which'].capitalize())
        if kwargs['which'] == 'projects':
            params = {k: kwargs[k] for k in {'project', 'sortkey', 'limit'}}
            self.analyze_logger.info('analyze projects params=%s', params)
            result = self._analyze_projects(project=p,
                                            sortkey=kwargs['sortkey'],
                                            limit=kwargs['limit'])
        elif kwargs['which'] == 'models':
            params = {k: kwargs[k] for k in {'project',
                                             'model',
                                             'timeframe',
                                             'min_queries',
                                             'sortkey',
                                             'limit'}}
            self.analyze_logger.info('analyze models params=%s', params)
            result = self._analyze_models(project=p,
                                          model=m,
                                          sortkey=kwargs['sortkey'],
                                          limit=kwargs['limit'],
                                          timeframe=kwargs['timeframe'],
                                          min_queries=kwargs['min_queries'])
        elif kwargs['which'] == 'explores':
            params = {k: kwargs[k] for k in {'model',
                                             'explore',
                                             'timeframe',
                                             'min_queries',
                                             'sortkey',
                                             'limit'}}
            self.analyze_logger.info('analyze explores params=%s', )
            result = self._analyze_explores(model=m,
                                            explore=kwargs['explore'],
                                            sortkey=kwargs['sortkey'],
                                            limit=kwargs['limit'],
                                            timeframe=kwargs['timeframe'],
                                            min_queries=kwargs['min_queries'])
        self.analyze_logger.info('Analyze Complete')

        result = tabulate(result, headers=headers,
                          tablefmt=format, numalign='center')

        return result

    def _analyze_projects(self, project=None, sortkey=None, limit=None):
        projects = fetcher.get_project_files(self, project=project)
        info = []
        for p in projects:
            metadata = list(map(lambda x:
                                'model' if x['type'] == 'model' else
                                ('view' if x['type'] == 'view' else None),
                                p['files']))

            model_count = metadata.count('model')
            view_count = metadata.count('view')
            git_tests = fetcher.test_git_connection(self, p['name'])
            info.append({
                'project': p['name'],
                'model_count': model_count,
                'view_count': view_count,
                'git_connection_status': git_tests,
                'pull_request_mode': p['pr_mode'],
                'validation_required': p['validation_required']
            })

        valid_values = list(info[0].keys())
        info = styler.sort(info, valid_values, sortkey)
        info = styler.limit(info, limit=limit)

        return info

    def _analyze_models(self, project=None, model=None,
                        sortkey=None, limit=None,
                        timeframe=90, min_queries=0):
        models = fetcher.get_models(self, project=project,
                                    model=model, verbose=1)
        used_models = fetcher.get_used_models(self, timeframe, min_queries)
        info = []
        for m in models:
            explore_count = len(m['explores'])
            if m['name'] in used_models:
                query_run_count = used_models[m['name']]
            else:
                query_run_count = 0
            unused_explores = fetcher.get_unused_explores(self, m['name'],
                                                          timeframe,
                                                          min_queries)
            info.append({
                'project': m['project_name'],
                'model': m['name'],
                'explore_count': explore_count,
                'unused_explores': len(unused_explores),
                'query_run_count': query_run_count
            })
        valid_values = list(info[0].keys())
        info = styler.sort(info, valid_values, sortkey)
        info = styler.limit(info, limit=limit)
        return info

    def _analyze_explores(self, model=None, explore=None,
                          sortkey=None, limit=None,
                          min_queries=0, timeframe=90):
        explores = fetcher.get_explores(self, model=model,
                                        explore=explore, verbose=1)
        explores_usage = {}
        info = []
        for e in explores:
            # in case explore does not exist (bug - #32748)
            if e is None:
                pass
            else:
                _used_fields = fetcher.get_used_explore_fields(self,
                                                               e['model_name'],
                                                               e['scopes'],
                                                               timeframe,
                                                               min_queries)
                used_fields = list(_used_fields.keys())
                exposed_fields = fetcher.get_explore_fields(self,
                                                            explore=e,
                                                            scoped_names=1)
                unused_fields = set(exposed_fields) - set(used_fields)
                field_count = len(exposed_fields)
                query_count = fetcher.get_used_explores(self,
                                                        model=e['model_name'],
                                                        explore=e['name'])

                all_joins = set(e['scopes'])
                all_joins.remove(e['name'])
                used_joins = set([i.split('.')[2] for i in used_fields])
                unused_joins = len(list(all_joins - used_joins))

                has_description = 'Yes' if e['description'] else 'No'

                if query_count.get(e['name']):
                    query_count = query_count[e['name']]
                else:
                    query_count = 0
                info.append({
                    'model': e['model_name'],
                    'explore': e['name'],
                    'is_hidden': e['hidden'],
                    'has_description': has_description,
                    'join_count': len(all_joins),
                    'unused_joins': unused_joins,
                    'field_count': field_count,
                    'unused_fields': len(unused_fields),
                    'query_count': query_count
                })

        if not info:
            self.analyze_logger.error('No matching explores found')
            raise Exception('No matching explores found')
        valid_values = list(info[0].keys())
        info = styler.sort(info, valid_values, sortkey)
        info = styler.limit(info, limit=limit)
        return info



#!/usr/bin/env python3
import yaml
from .modules.lookerapi import LookerApi
from itertools import groupby
import argparse
import os
import errno
import sys
from operator import itemgetter
from .modules.spinner import Spinner
import threading
from tabulate import tabulate
from .modules.auth import authenticate
import logging.config
import henry
LOGGING_CONFIG_PATH = os.path.join(os.path.dirname(henry.__file__),
                                   '.support_files/logging.conf')
LOGGING_LOG_PATH = os.path.join(os.path.expanduser('~'), '.henry')
if not os.path.exists(LOGGING_LOG_PATH):
    os.mkdir(LOGGING_LOG_PATH)
elif os.path.exists(LOGGING_LOG_PATH) and not os.path.isdir(LOGGING_LOG_PATH):
    print('Cannot create log directory in %s' % LOGGING_LOG_PATH)
    sys.exit(1)
LOGGING_LOG_PATH = os.path.join(LOGGING_LOG_PATH, 'henry.log')
logging.config.fileConfig(LOGGING_CONFIG_PATH,
                          defaults={'logfilename': LOGGING_LOG_PATH},
                          disable_existing_loggers=False)
from .commands.analyze import Analyze
from .commands.vacuum import Vacuum
from .commands.pulse import Pulse

logger = logging.getLogger('main')
# sys.tracebacklimit = -1 # enable only on shipped release


def main():
    logger.info('Starting henry')
    HELP_PATH = os.path.join(os.path.dirname(henry.__file__),
                             '.support_files/help.rtf')
    with open(HELP_PATH, 'r', encoding='unicode_escape') as myfile:
        descStr = myfile.read()

    parser = argparse.ArgumentParser(
        description=descStr,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog='henry',
        usage='henry command subcommand '
              '[subcommand options] [global '
              'options]',
        allow_abbrev=False,
        add_help=False)

    subparsers = parser.add_subparsers(dest='command',
                                       help=argparse.SUPPRESS)
    parser.add_argument("-h", "--help", action="help", help=argparse.SUPPRESS)

    # subparsers.required = True # works, but might do without for now.

    pulse = subparsers.add_parser('pulse', help='pulse help')

    analyze_parser = subparsers.add_parser('analyze', help='analyze help',
                                           usage='henry analyze')
    analyze_parser.set_defaults(which=None)
    analyze_subparsers = analyze_parser.add_subparsers()
    analyze_projects = analyze_subparsers.add_parser('projects')
    analyze_models = analyze_subparsers.add_parser('models')
    analyze_explores = analyze_subparsers.add_parser('explores')

    # project subcommand
    analyze_projects.set_defaults(which='projects')
    analyze_projects.add_argument('-p', '--project',
                                  type=str,
                                  default=None,
                                  help='Filter on a project')
    analyze_projects.add_argument('--order_by',
                                  nargs=2,
                                  metavar=('ORDER_FIELD', 'ASC/DESC'),
                                  dest='sortkey',
                                  help='Sort results by a field')
    analyze_projects.add_argument('--limit',
                                  type=int,
                                  default=None,
                                  nargs=1,
                                  help='Limit results. No limit by default')

    # models subcommand
    analyze_models.set_defaults(which='models')
    models_group = analyze_models.add_mutually_exclusive_group()

    models_group.add_argument('-p', '--project',
                              type=str,
                              default=None,
                              help='Filter on project')
    models_group.add_argument('-model', '--model',
                              type=str,
                              default=None,
                              help='Filter on model')
    analyze_models.add_argument('--timeframe',
                                type=int,
                                default=90,
                                help='Timeframe (between 0 and 90)')
    analyze_models.add_argument('--min_queries',
                                type=int,
                                default=0,
                                help='Query threshold')
    analyze_models.add_argument('--order_by',
                                nargs=2,
                                metavar=('ORDER_FIELD', 'ASC/DESC'),
                                dest='sortkey',
                                help='Sort results by a field')
    analyze_models.add_argument('--limit',
                                type=int,
                                default=None,
                                nargs=1,
                                help='Limit results. No limit by default')

    # explores subcommand
    analyze_explores.set_defaults(which='explores')
    analyze_explores.add_argument('-model', '--model',
                                  type=str,
                                  default=None,
                                  required=('--explore') in sys.argv,
                                  help='Filter on model')
    analyze_explores.add_argument('-e', '--explore',
                                  default=None,
                                  help='Filter on model')
    analyze_explores.add_argument('--timeframe',
                                  type=int,
                                  default=90,
                                  help='Timeframe (between 0 and 90)')
    analyze_explores.add_argument('--min_queries',
                                  type=int,
                                  default=0,
                                  help='Query threshold')
    analyze_explores.add_argument('--order_by',
                                  nargs=2,
                                  metavar=('ORDER_FIELD', 'ASC/DESC'),
                                  dest='sortkey',
                                  help='Sort results by a field')
    analyze_explores.add_argument('--limit',
                                  type=int,
                                  default=None,
                                  nargs=1,
                                  help='Limit results. No limit by default')

    # VACUUM Subcommand
    vacuum_parser = subparsers.add_parser('vacuum', help='vacuum help',
                                          usage='henry vacuum')
    vacuum_parser.set_defaults(which=None)
    vacuum_subparsers = vacuum_parser.add_subparsers()
    vacuum_models = vacuum_subparsers.add_parser('models')
    vacuum_explores = vacuum_subparsers.add_parser('explores')
    vacuum_models.set_defaults(which='models')
    vm_group = vacuum_models.add_mutually_exclusive_group()
    vm_group.add_argument('-p', '--project',
                          type=str,
                          default=None,
                          help='Filter on Project')
    vm_group.add_argument('-m', '--model',
                          type=str,
                          default=None,
                          help='Filter on model')

    vacuum_models.add_argument('--timeframe',
                               type=int,
                               default=90,
                               help='Usage period to examine (in the range of '
                                    '0-90 days). Default: 90 days.')

    vacuum_models.add_argument('--min_queries',
                               type=int,
                               default=0,
                               help='Vacuum threshold. Explores with less '
                                    'queries in the given usage period will '
                                    'be vacuumed. Default: 0 queries.')

    vacuum_explores.set_defaults(which='explores')
    vacuum_explores.add_argument('-m', '--model',
                                 type=str,
                                 default=None,
                                 required=('--explore') in sys.argv,
                                 help='Filter on model')

    vacuum_explores.add_argument('-e', '--explore',
                                 type=str,
                                 default=None,
                                 help='Filter on explore')

    vacuum_explores.add_argument('--timeframe',
                                 type=int,
                                 default=90,
                                 help='Timeframe (between 0 and 90)')

    vacuum_explores.add_argument('--min_queries',
                                 type=int,
                                 default=0,
                                 help='Query threshold')

    for subparser in [analyze_projects, analyze_models, analyze_explores,
                      vacuum_models, vacuum_explores, pulse]:
        subparser.add_argument('--output',
                               type=str,
                               default=None,
                               help='Path to file for saving the output')
        subparser.add_argument('-q', '--quiet',
                               action='store_true',
                               help='Silence output')
        subparser.add_argument('--plain',
                               default=None,
                               action='store_true',
                               help='Show results in a table format '
                                    'without the gridlines')
        subparser.add_argument_group("Authentication")
        subparser.add_argument('--host', type=str, default='looker',
                               required=any(k in sys.argv for k in
                                            ['--client_id', '--cliet_secret',
                                             '--alias']),
                               help=argparse.SUPPRESS)
        subparser.add_argument('--port', type=int, default=19999,
                               help=argparse.SUPPRESS)
        subparser.add_argument('--client_id', type=str,
                               required=any(k in sys.argv for k in
                                            ['--client_secret', '--alias']),
                               help=argparse.SUPPRESS)
        subparser.add_argument('--client_secret', type=str,
                               required=any(k in sys.argv for k in
                                            ['--client_id', '--alias']),
                               help=argparse.SUPPRESS)
        subparser.add_argument('--persist', action='store_true',
                               help=argparse.SUPPRESS)
        subparser.add_argument('--alias', type=str,
                               help=argparse.SUPPRESS)
        subparser.add_argument('--path', type=str, default='',
                               help=argparse.SUPPRESS)

    args = vars(parser.parse_args())
    _args = {}
    for key, value in args.items():
        if key == 'client_secret':
            _args[key] = '[FILTERED]'
        else:
            _args[key] = value
    logger.info('Parsing args, %s', _args)

    if not args['command']:
        print('usage:', parser.usage)
        print('\nNo command specified. Try `henry --help` for help.')
        sys.exit(1)
    auth_params = ('host', 'port', 'client_id', 'client_secret', 'persist',
                   'alias', 'path')
    auth_args = {k: args[k] for k in auth_params}

    # authenticate
    looker = authenticate(**auth_args)

    # map subcommand to function
    if args['command'] in ('analyze', 'vacuum'):
        if args['which'] is None:
            parser.error("No command")
        else:
            with Spinner():
                if args['command'] == 'analyze':
                    analyze = Analyze(looker)
                    result = analyze.analyze(**args)
                else:
                    vacuum = Vacuum(looker)
                    result = vacuum.vacuum(**args)
        # silence outout if --silence flag is used
        if not args['quiet']:
            print(result)
    elif args['command'] == 'pulse':
                pulse = Pulse(looker)
                result = pulse.run_all()
    else:
        print('No command passed')

    # save to file if --output flag is used
    if args['output']:
        logger.info('Saving results to file: %s', args['output'])
        if os.path.isdir(args['output']):
            error = IsADirectoryError(errno.EISDIR,
                                      os.strerror(errno.EISDIR),
                                      args['output'])
            logger.error(error)
            raise error
        elif not args['output'].endswith('.txt'):
            error = ValueError('Output file must be a .txt file')
            logger.exception(error)
            raise error
        elif os.path.isfile(args['output']):
            error = FileExistsError(errno.EEXIST,
                                    os.strerror(errno.EEXIST),
                                    args['output'])
            logger.error(error)
            raise error
        else:
            try:
                f = open(args['output'], 'w+')
                f.write(result+'\n')
                f.close()
                logger.info('Results succesfully saved.')
            except Exception as e:
                logger.error(e)
                raise(e)


if __name__ == "__main__":
    main()



import os
import logging
import yaml
import sys
from .lookerapi import LookerApi

auth_logger = logging.getLogger('auth')


# returns an instanstiated Looker object using the
# credentials supplied by the auth argument group
def authenticate(**kwargs):
    auth_logger.info('Authenticating into Looker API')
    filepath = kwargs['path'] + 'config.yml'
    cleanpath = os.path.abspath(filepath)
    if kwargs['client_id'] and kwargs['client_secret']:
        auth_logger.info('Fetching auth params passed in CLI')
        host = kwargs['host']
        client_id = kwargs['client_id']
        client_secret = kwargs['client_secret']
        token = None
    else:
        auth_logger.info('Checking permissions for %s', cleanpath)
        st = os.stat(cleanpath)
        ap = oct(st.st_mode)
        if ap != '0o100600':
            print('Config file permissions are set to %s and are not strict '
                  'enough. Change to rw------- or 600 and try again.' % ap)
            auth_logger.warning('Config file permissions are %s and not strict'
                                ' enough.' % ap)
            sys.exit(1)
        auth_logger.info('Opening config file from %s' % cleanpath)
        try:
            f = open(cleanpath, 'r')
            params = yaml.load(f)
            f.close()
        except FileNotFoundError as error:
            auth_logger.exception(error, exc_info=False)
            print('ERROR: %s not found' % filepath)
            sys.exit(1)

        try:
            auth_logger.info('Fetching auth credentials from file, %s',
                             cleanpath)
            host = params['hosts'][kwargs['host']]['host']
            client_secret = params['hosts'][kwargs['host']]['secret']
            client_id = params['hosts'][kwargs['host']]['id']
            #  last auth token. Works if --persist was previously used,
            # otherwise it fails)
            token = params['hosts'][kwargs['host']]['access_token']
        except KeyError as error:
            auth_logger.error('Auth Error: %s not found' % error,
                              exc_info=False)
            print('ERROR: %s not found' % error)
            sys.exit(1)

    auth_logger.info('auth params=%s', {'host': host,
                                        'port': kwargs['port'],
                                        'client_id': client_id,
                                        'client_secret': "[FILTERED]"})
    looker = LookerApi(host=host,
                       port=kwargs['port'],
                       id=client_id,
                       secret=client_secret,
                       access_token=token)
    auth_logger.info('Authentication Successful')

    if kwargs['alias']:
        auth_logger.info('Saving credentials to file: %s', cleanpath)
        with open(cleanpath, 'r') as f:
            params = yaml.safe_load(f)
            params['hosts'][kwargs['alias']] = {}
            params['hosts'][kwargs['alias']]['host'] = host
            params['hosts'][kwargs['alias']]['id'] = client_id
            params['hosts'][kwargs['alias']]['secret'] = client_secret
            params['hosts'][kwargs['alias']]['access_token'] = ''

        with open(cleanpath, 'w') as f:
            yaml.safe_dump(params, f, default_flow_style=False)

        os.chmod(cleanpath, 0o600)

    if kwargs['persist']:
        auth_logger.info('Persisting API session. Saving auth token under '
                         '%s in %s', host, cleanpath)
        with open(cleanpath, 'r+') as f:
            params = yaml.safe_load(f)
            access_token = looker.get_access_token()
            params['hosts'][kwargs['host']]['access_token'] = access_token

        with open(cleanpath, 'w') as f:
            yaml.safe_dump(params, f, default_flow_style=False)

        os.chmod(cleanpath, 0o600)

    return looker



import logging
import re
import requests
from textwrap import fill
from tqdm import tqdm
from tabulate import tabulate
from tqdm import trange
from henry.modules.color import color


class Pulse(object):

    postfix_default = [dict(value="RUNNING")]

    def __init__(self, looker):
        self.looker = looker
        self.pulse_logger = logging.getLogger('pulse')
        self.bar = '%s%s{postfix[0][value]}%s {desc}: ' \
                   '{percentage:3.0f}%% |{bar}|[{elapsed}<' \
                   '{remaining}]' % (color.BOLD,
                                     color.GREEN,
                                     color.ENDC)
        self.postfix_default = [dict(value="RUNNING")]

    def run_all(self):
        self.pulse_logger.info('Checking instance pulse')
        self.pulse_logger.info('Checking Connections')
        result = self.check_connections()
        print(result, end='\n\n')
        self.pulse_logger.info('Complete: Checking Connections')

        self.pulse_logger.info('Analyzing Query Stats')
        r1, r2, r3 = self.check_query_stats()
        print(r1)
        print(r2)
        print(r3, end='\n\n')
        self.pulse_logger.info('Complete: Analyzing Query Stats')

        # check scheduled plans
        self.pulse_logger.info('Analyzing Query Stats')
        with trange(1, desc='(3/5) Analyzing Scheduled Plans',
                    bar_format=self.bar, postfix=self.postfix_default,
                    ncols=100, miniters=0) as t:
            for i in t:
                result = self.check_scheduled_plans()
                fail_flag = 0
                if type(result) == list and len(result) > 0:
                    if result[0]['failure'] > 0:
                        fail_flag = 1
                    result = tabulate(result, headers="keys",
                                      tablefmt='psql', numalign='center')
                t.postfix[0]["value"] = 'DONE'
                t.update()
        print(result, end='\n\n')
        if fail_flag == 1:
            print('Navigate to /admin/scheduled_jobs on your instance for '
                  'more details', end='\n\n')
        self.pulse_logger.info('Complete: Analyzing Scheduled Plans')

        # check enabled legacy features
        self.pulse_logger.info('Checking Legacy Features')
        with trange(1, desc='(4/5) Legacy Features', bar_format=self.bar,
                    postfix=self.postfix_default, ncols=100, miniters=0) as t:
            for i in t:
                result = self.check_legacy_features()
                t.postfix[0]["value"] = 'DONE'
                t.update()
        print(result, end='\n\n')
        self.pulse_logger.info('Complete: Checking Legacy Features')

        # check looker version
        self.pulse_logger.info('Checking Version')
        t = trange(1, desc='(5/5) Version', bar_format=self.bar,
                   postfix=self.postfix_default, ncols=100)
        for i in t:
            result = self.check_version()
            t.postfix[0]["value"] = "DONE"
            t.update()
        print(result, end='\n\n')
        self.pulse_logger.info('Complete: Checking Version')
        self.pulse_logger.info('Complete: Checking instance pulse')

        return

    def check_connections(self):
        result = []
        connections = []
        for c in self.looker.get_connections():
            if c['name'] != 'looker':
                c_tests = (', ').join(c['dialect']['connection_tests'])
                c_name = c['name']
                connections.append((c_name, c_tests))

        with tqdm(total=len(connections), desc='(1/5) Testing Connections',
                  bar_format=self.bar, postfix=self.postfix_default,
                  ncols=100, miniters=0) as t:
            for idx, (c, tests) in enumerate(connections):
                tests = {'tests': tests}
                results = self.looker.test_connection(c, tests)
                formatted_results = []
                fail_flag = 0
                for i in results:
                    if i['status'] == 'error':
                        formatted_results.append('-- ' + fill(i['message'],
                                                              width=100))
                        fail_flag = 1
                formatted_results = list(set(formatted_results))
                status = '\n'.join(formatted_results)
                result.append({'Connection': c,
                               'Status': 'OK' if fail_flag == 0 else status})
                if idx == len(connections) - 1:
                    t.postfix[0]['value'] = 'DONE'
                t.update()

        return tabulate(result, headers="keys", tablefmt='psql')

    def check_query_stats(self):
        # check query stats
        with trange(3, desc='(2/5) Analyzing Query Stats', bar_format=self.bar,
                    postfix=self.postfix_default, ncols=100, miniters=0) as t:
            for i in t:
                if i == 0:
                    query_count = self.get_query_type_count()
                if i == 1:
                    query_runtime_stats = self.get_query_stats('complete')
                if i == 2:
                    slow_queries = self.get_slow_queries(
                                                 query_runtime_stats['avg']*5)
                    t.postfix[0]['value'] = 'DONE'

        r1 = '{} queries run, ' \
             '{} errored, ' \
             '{} killed'.format(query_count['total'], query_count['errored'],
                                query_count['killed'])
        r2 = 'Query Runtime min/avg/max: ' \
             '{}/{}/{} seconds'.format(query_runtime_stats['min'],
                                       query_runtime_stats['avg'],
                                       query_runtime_stats['max'])

        if slow_queries:
            r3 = 'Query IDs for queries that took more than 5x the average ' \
                 'query runtime : {}'.format(slow_queries)
            r3 = fill(r3, width=80)
        else:
            r3 = 'No abnormally slow queries found'
        return r1, r2, r3

    # get number of queries run, killed, completed, errored, queued
    def get_slow_queries(self, avg_runtime):
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": [
                "query.id",
            ],
            "filters": {
                "query.id": "NOT NULL",
                "history.created_date": "30 days",
                "history.status": "-NULL",
                "history.result_source": "query",
                "query.model": "-i^_^_looker",
                "history.total_runtime": ">=" + str(avg_runtime)
            },
            "sorts": [
                "query.id asc"
            ],
            "limit": "50000"
        }

        r = self.looker.run_inline_query(result_format="json", body=body,
                                         fields={"cache": "false"})

        if r:
            ids = (', ').join([str(query['query.id']) for query in r])
        else:
            ids = None
        return ids

    # get number of queries run, killed, completed, errored, queued
    def get_query_type_count(self):
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": [
                "history.query_run_count",
                "history.status",
                "history.created_date"
            ],
            "pivots": [
                "history.status"
            ],
            "filters": {
                "history.created_date": "30 days",
                "history.status": "-NULL",
                "history.result_source": "query",
                "query.model": "-i^_^_looker"
            },
            "sorts": [
                "history.created_date desc",
                "history.result_source"
            ],
            "limit": "50000"
        }

        r = self.looker.run_inline_query(result_format="json", body=body,
                                         fields={"cache": "false"})
        completed = 0
        errored = 0
        killed = 0
        queued = 0
        if(len(r) > 0):
            for entry in r:
                e = entry['history.query_run_count']['history.status']
                if 'complete' in e:
                    c_i = e['complete']
                else:
                    c_i = 0
                c_i = c_i if c_i is not None else 0
                completed += c_i

                if 'error' in e:
                    e_i = e['error']
                else:
                    e_i = 0
                e_i = e_i if e_i is not None else 0
                errored += e_i

                if 'killed' in e:
                    k_i = e['killed']
                else:
                    k_i = 0
                k_i = k_i if k_i is not None else 0
                killed += k_i

                if 'pending' in e:
                    q_i = e['pending']
                else:
                    q_i = 0
                q_i = q_i if q_i is not None else 0
                queued += q_i

        response = {'total': completed + errored + killed,
                    'completed': completed,
                    'errored': errored,
                    'killed': killed,
                    'queued': queued}

        return response

    # get number of queries run, killed, completed, errored, queued
    def get_query_stats(self, status):
        valid_statuses = ['error', 'complete', 'running']
        if status not in valid_statuses:
            raise ValueError("Invalid query status, must be in %r"
                             % valid_statuses)
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": [
                "history.min_runtime",
                "history.max_runtime",
                "history.average_runtime",
                "history.total_runtime"
            ],
            "filters": {
                "history.created_date": "30 days",
                "history.status": status,
                "query.model": "-i^_^_looker"
            },
            "limit": "50000"
        }

        r = self.looker.run_inline_query(result_format="json", body=body,
                                         fields={"cache": "false"})[0]
        for i in ('history.min_runtime',
                  'history.max_runtime',
                  'history.average_runtime'):
            if r[i] is not None:
                r[i] = round(r[i], 2)
            else:
                r[i] = '-'
        response = {'min': r['history.min_runtime'],
                    'max': r['history.max_runtime'],
                    'avg': r['history.average_runtime'],
                    'total': r['history.total_runtime']}

        return response

    def check_scheduled_plans(self):
        body = {
            "model": "i__looker",
            "view": "scheduled_plan",
            "fields": ["scheduled_job.status", "scheduled_job.count"],
            "pivots": ["scheduled_job.status"],
            "filters": {
                    "scheduled_plan.run_once": "no",
                    "scheduled_job.status": "-NULL",
                    "scheduled_job.created_date": "30 days"
            },
            "limit": "50000"
        }

        r = self.looker.run_inline_query("json", body)
        result = []
        if r:
            r = r[0]['scheduled_job.count']['scheduled_job.status']
            failed = r['failure'] or 0
            succeeded = r['success'] or 0
            result.append({'total': failed + succeeded,
                           'failure': failed,
                           'success': succeeded})
            return result
        else:
            return "No Plans Found"

    def check_integrations(self):
        response = self.looker.get_integrations()
        integrations = []
        for r in response:
            if r['enabled']:
                integrations.append(r['label'])

        result = None if len(integrations) == 0 else integrations

        return result

    def check_legacy_features(self):
        response = self.looker.get_legacy_features()
        _result = []
        for r in response:
            if r['enabled'] is True:
                _result.append({'Legacy Features': r['name']})

        if _result:
            result = tabulate(_result, headers="keys", tablefmt='psql')
        else:
            result = 'No legacy features found'
        return result

    def check_version(self):
        _v = self.looker.get_version()['looker_release_version']
        version = re.findall(r'(\d.\d+)', _v)[0]
        session = requests.Session()
        _lv = session.get('https://learn.looker.com:19999/versions').json()
        _lv = _lv['looker_release_version']
        latest_version = re.findall(r'(\d.\d+)', _lv)[0]
        if version == latest_version:
            result = '{} ({})'.format(version, 'up-to-date')
        else:
            result = '{} ({} {})'.format(version,
                                         'outdated, latest version is',
                                         latest_version)
        return result



# color.py


# string is the original string
# type determines the color/label added
# style can be color or text
class color(object):
    GREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'

    def format(self, string, type, style='color'):
        formatted_string = ''
        string = str(string)
        if style == 'color':
            if type in ('success', 'pass'):
                formatted_string += self.GREEN + string + self.ENDC
            elif type == 'warning':
                formatted_string += self.WARNING + string + self.ENDC
            elif type in ('error', 'fail'):
                formatted_string += self.FAIL + string + self.ENDC
        elif style == 'text':
            if type == 'success':
                formatted_string += string
            elif type == 'warning':
                formatted_string += 'WARNING: ' + string
            elif type == 'error':
                formatted_string += 'ERROR: ' + string

        return formatted_string



from . import styler
from collections import Counter
import logging
import re
from textwrap import fill


class Fetcher(object):
    def __init__(self, looker):
        self.looker = looker
        self.fetch_logger = logging.getLogger('fetcher')

    def get_project_files(self, project=None):
        self.fetch_logger.info('Fetching projects, %s', locals())
        if project is None:
            self.fetch_logger.info('Fetching all project files')
            projects = self.looker.get_projects()
        else:
            self.fetch_logger.info('Fetching project files for %s', project)
            projects = self.looker.get_project(project)

        project_data = []
        for p in projects:
            project_files = self.looker.get_project_files(project=p['id'])

            project_data.append({
                'name': p['id'],
                'pr_mode': p['pull_request_mode'],
                'validation_required': p['validation_required'],
                'git_remote_url': p['git_remote_url'],
                'files': project_files
            })
        self.fetch_logger.info('Fetch Complete :: Projects')
        return project_data

    # function that returns list of model definitions or model names (with
    # verbose 0 or 1 respectively) Allows the user to specify a project name,
    # a model name or nothing at all. project paramater is a string while model
    # parameter is a list
    def get_models(self, project=None, model=None, verbose=0, scoped_names=0):
        if project is None and model is None:
            self.fetch_logger.info('Fetching all models, %s', locals())
            models = self.looker.get_models()
        elif project is not None and model is None:
            # if no parameters are specified
            self.fetch_logger.info('Fetching all models in %s, %s', project,
                                   locals())
            r = self.looker.get_models()
            models = list(filter(lambda x: x['project_name'] == project, r))
            if not models:
                self.fetch_logger.error('Project not found')
                raise Exception('Project not found')
        elif project is not None and model is not None:
            # if both project and model paramaters are specified
            self.fetch_logger.info('Warning: Project parameter ignored. \
                                   Model names are unique across projects.')
            models = [self.looker.get_model(model)]
        else:
            # if project parameter wasn't passed but model was.
            self.fetch_logger.info('Fetching model %s, %s', model, locals())
            models = self.looker.get_model(model_name=model)

        models = list(filter(lambda x: x['has_content'] is True, models))
        if verbose == 0:
            models = [(m['project_name'] + ".") * scoped_names + m['name']
                      for m in models]
        self.fetch_logger.info('Fetch Complete :: Models')
        return models

    def get_used_models(self, timeframe=90, min_queries=0):
        self.fetch_logger.info('Fetching used models from i__looker, %s',
                               locals())
        timeframe = str(timeframe) + ' days'
        min_queries = '>=' + str(min_queries)
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": ["query.model", "history.query_run_count"],
            "filters": {"history.created_date": timeframe,
                        "query.model": "-i^_^_looker",
                        "history.query_run_count": min_queries
                        },
            "limit": "50000"
        }

        response = self.looker.run_inline_query("json", body)

        x = {}
        for r in response:
            x[r['query.model']] = r['history.query_run_count']
        self.fetch_logger.info('Used Models Fetch Complete')
        return(x)

    # errors have to be handled more downstream if explore does not exist due
    # to bug #32748
    def get_explores(self, model=None, explore=None, scoped_names=0,
                     verbose=0):
        explores = []
        if explore is not None:
            self.fetch_logger.info('Fetching explore %s, %s', explore,
                                   locals())
            e = self.looker.get_explore(model_name=model, explore_name=explore)
            if e:
                explores.extend(e)
        else:
            self.fetch_logger.info('Fetching all explores, %s', locals())
            models = self.get_models(model=model, verbose=1)
            for mdl in models:
                for e in mdl['explores']:
                    if verbose == 1:
                        explores.extend(self.looker.get_explore(mdl['name'],
                                                                e['name']))
                    else:
                        explores.append((mdl['name'], e['name']))
        self.fetch_logger.info('Fetch Complete :: Explores')
        return explores

    def get_explore_fields(self, explore=None, scoped_names=0):
        self.fetch_logger.info('Parsing explore body for exposed fields')
        fields = []
        for dimension in explore['fields']['dimensions']:
            if dimension['hidden'] is not True:
                fields.append((explore['model_name']+'.'
                              + explore['name'] + '.')*scoped_names
                              + dimension['name'])
        for measure in explore['fields']['measures']:
            if measure['hidden'] is not True:
                fields.append((explore['model_name']+'.'
                              + explore['name']+'.')*scoped_names
                              + measure['name'])
        for fltr in explore['fields']['filters']:
            if fltr['hidden'] is not True:
                fields.append((explore['model_name']+'.'
                              + explore['name']+'.')*scoped_names
                              + fltr['name'])
        self.fetch_logger.info('Parsing Complete')
        return list(set(fields))

    def get_unused_explores(self, model=None, timeframe=90, min_queries=0):
        self.fetch_logger.info('Fetching unused explores, %s', locals())
        used_explores = self.get_used_explores(model=model,
                                               timeframe=timeframe,
                                               min_queries=min_queries)
        used_explores = used_explores.keys()
        all_explores = self.get_explores(model=model)
        all_explores = [i[1] for i in all_explores]
        unused_explores = list(set(all_explores) - set(used_explores))
        self.fetch_logger.info('Fetch Complete:: Unused Explores')
        return unused_explores

    # function that runs i__looker query and returns fully scoped fields used
    # remember explore names are not unique, filter on model as well
    # query.explore is the actual explore name
    # query.model is the model
    # query.fields are is view.field (view is view name used in the explore)
    # to uniquely identify fields, explore.view.field should be used,
    # or even better, model.explore.view.field
    def get_used_explore_fields(self, model=None, explore=None, timeframe=90,
                                min_queries=0):
        self.fetch_logger.info('Fetching exposed explore fields, %s', locals())
        m = model.replace('_', '^_') + ',' if model is not None else ''
        m += "-i^_^_looker"
        e = ','.join(explore).replace('_', '^_')
        min_queries = '>=' + str(min_queries)
        timeframe = str(timeframe) + ' days'
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": ["query.model", "query.view",
                       "query.formatted_fields",
                       "query.formatted_filters", "query.sorts",
                       "query.formatted_pivots",
                       "history.query_run_count"],
            "filters": {"history.created_date": timeframe,
                        "query.model": m,
                        "query.view": e,
                        "history.query_run_count": min_queries},
            "limit": "50000"
        }
        # returns only fields used from a given explore
        response = self.looker.run_inline_query("json", body)

        formatted_fields = []
        for row in response:
            fields = []
            explore = row['query.view']
            model = row['query.model']
            run_count = row['history.query_run_count']
            fields.extend(re.findall(r'(\w+\.\w+)',
                                     str(row['query.formatted_fields'])))
            fields.extend(re.findall(r'(\w+\.\w+)',
                                     str(row['query.formatted_filters'])))
            fields.extend(re.findall(r'(\w+\.\w+)',
                                     str(row['query.formatted_pivots'])))
            fields.extend(re.findall(r'(\w+\.\w+)',
                                     str(row['query.sorts'])))
            for field in fields:
                s = model + '.' + explore + '.' + field + '.' + str(run_count)
                formatted_fields.extend([s])

        field_name = []
        field_use_count = []
        for row in formatted_fields:
            field = '.'.join(row.split('.')[:-1])  # remove the count
            field_name.append(field)  # model.explore.view scoped
            count = int(row.split('.')[-1])
            field_use_count.append({
                'field_name': field,
                'count': count
            })

        c = Counter()

        for value in field_use_count:
            c[value['field_name']] += value['count']
        self.fetch_logger.info('Fetch Complete :: Exposed Explore Fields ')
        return dict(c)

    def get_used_explores(self, model=None, explore=None,
                          timeframe=90, min_queries=0):
        self.fetch_logger.info('Fetching used explores, %s', locals())
        timeframe = str(timeframe) + ' days'
        min_queries = '>=' + str(min_queries)
        m = model.replace('_', '^_') + ',' if model is not None else ''
        body = {
            "model": "i__looker",
            "view": "history",
            "fields": ["query.view", "history.query_run_count"],
            "filters": {"history.created_date": timeframe,
                        "query.model": m,
                        "history.query_run_count": min_queries,
                        "query.view": explore
                        },
            "limit": "50000"
        }

        response = self.looker.run_inline_query("json", body)

        x = {}
        for r in response:
            x[r['query.view']] = r['history.query_run_count']
        self.fetch_logger.info('Fetch Complete :: Used Explores')
        return(x)

    def test_git_connection(self, project):
        # enter dev mode
        self.looker.update_session(mode='dev')
        # obtain tests available
        tests = []
        for test in self.looker.git_connection_tests(project_id=project):
            tests.append(test['id'])
        verbose_result = []
        fail_flag = 0
        for idx, test in enumerate(tests):
            r = self.looker.run_git_connection_test(project_id=project,
                                                    test_id=test)
            verbose_result.append(r['id'] + ' (' + r['status'] + ')')
            if r['status'] != 'pass':
                fail_flag = 1
        verbose_result = ('\n').join(verbose_result)
        result = verbose_result if fail_flag == 1 else 'OK'
        return result



# styler.py
import logging
from tabulate import tabulate
from operator import itemgetter

style_logger = logging.getLogger('styler')


def limit(data, limit=None):
    if limit is not None:
        style_logger.info('Limiting results to %s', limit[0])
        return data[:limit[0]]
    else:
        return data


def sort(data, valid_values, sortkey):
    if sortkey is None:
        return data
    else:
        style_logger.info('Sort params=> %s', sortkey)
        valid_types = {'ASC': False, 'DESC': True}
        if sortkey[1].upper() in valid_types.keys():
            type = valid_types[sortkey[1].upper()]
        else:
            type = None

        sk = sortkey[0] if sortkey[0] in valid_values else False
        if not sk:
            style_logger.error('Sortkey:%s is invalid', sortkey[0])
            raise ValueError('Unrecognised order_by field, must be in %r' %
                             valid_values)
        elif type is None:
            style_logger.error('Sort type is invalid')
            raise ValueError('Unrecognised order_by field, must be in %r' %
                             list(valid_types.keys()))
        else:
            style_logger.info('Sorting data by %s %s', sk, type)
            data = sorted(data, key=itemgetter(sk), reverse=type)
    return data



import logging
from henry.modules import styler
from henry.modules.fetcher import Fetcher as fetcher
import re


class Vacuum(fetcher):
    def __init__(self, looker):
        super().__init__(looker)
        self.vacuum_logger = logging.getLogger('vacuum')

    def vacuum(self, **kwargs):
        p = kwargs['project'] if 'project' in kwargs.keys() else None
        m = kwargs['model'] if 'model' in kwargs.keys() else None
        format = 'plain' if kwargs['plain'] else 'psql'
        headers = '' if kwargs['plain'] else 'keys'
        if kwargs['which'] == 'models':
            self.vacuum_logger.info('Vacuuming Models')
            params = {k: kwargs[k] for k in {'project',
                                             'model',
                                             'timeframe',
                                             'min_queries'}}
            self.vacuum_logger.info('vacuum models params=%s', params)
            result = self._vacuum_models(project=p,
                                         model=m,
                                         min_queries=kwargs['min_queries'],
                                         timeframe=kwargs['timeframe'])
        if kwargs['which'] == 'explores':
            self.vacuum_logger.info('Vacuuming Explores')
            params = {k: kwargs[k] for k in {'model',
                                             'explore',
                                             'timeframe',
                                             'min_queries'}}
            self.vacuum_logger.info('vacuum explores params=%s', params),
            result = self._vacuum_explores(model=m,
                                           explore=kwargs['explore'],
                                           min_queries=kwargs['min_queries'],
                                           timeframe=kwargs['timeframe'])
        self.vacuum_logger.info('Vacuum Complete')
        result = styler.tabulate(result, headers=headers,
                                 tablefmt=format, numalign='center')
        return result

    def _vacuum_models(self, project=None, model=None, timeframe=90,
                       min_queries=0):
        if model is None:
            model = fetcher.get_models(self, project=project)
        else:
            model = model.split()
        used_models = fetcher.get_used_models(self, timeframe)
        info = []
        for m in model:
            explores = [e['name'] for e in fetcher.get_explores(self, model=m,
                                                                verbose=1)]
            unused_explores = fetcher.get_unused_explores(self, m,
                                                          timeframe,
                                                          min_queries)
            query_run_count = used_models[m] if m in used_models.keys() else 0
            unused_explores = ('\n').join(unused_explores)
            info.append({
                        'model': m,
                        'unused_explores': unused_explores or 'None',
                        'model_query_run_count': query_run_count})

        return info

    def _vacuum_explores(self, model=None, explore=None, timeframe=90,
                         min_queries=0):
        explores = fetcher.get_explores(self,
                                        model=model,
                                        explore=explore,
                                        verbose=1)
        info = []
        for e in explores:
            # get field usage from i__looker using all the views inside explore
            # returns fields in the form of model.explore.view.field
            _used_fields = fetcher.get_used_explore_fields(self,
                                                           e['model_name'],
                                                           e['scopes'],
                                                           timeframe,
                                                           min_queries)
            used_fields = list(_used_fields.keys())
            # get field picker fields in the form of model.explore.view.field
            exposed_fields = fetcher.get_explore_fields(self,
                                                        explore=e,
                                                        scoped_names=1)
            _unused_fields = set(exposed_fields) - set(used_fields)

            # remove scoping
            all_joins = set(e['scopes'])
            all_joins.remove(e['name'])
            used_joins = set([i.split('.')[2] for i in used_fields])

            _unused_joins = list(all_joins - used_joins)
            unused_joins = ('\n').join(_unused_joins) or 'N/A'

            # only keep fields that belong to used joins (unused joins fields
            # don't matter) if there's at least one used join (including the
            # base view). else don't match anything
            temp = list(used_joins)
            temp.append(e['name'])
            pattern = ('|').join(temp) or 'ALL'
            unused_fields = []
            if pattern != 'ALL':
                for field in _unused_fields:
                    f = re.match(r'^({0}).*'.format(pattern),
                                 '.'.join(field.split('.')[2:]))
                    if f is not None:
                        unused_fields.append(f.group(0))
                unused_fields = sorted(unused_fields)
                unused_fields = ('\n').join(unused_fields)
            else:
                unused_fields = styler.color.format(pattern,
                                                    'fail',
                                                    'color')
            info.append({
                        'model': e['model_name'],
                        'explore': e['name'],
                        'unused_joins': unused_joins,
                        'unused_fields': unused_fields
                        })
        if not info:
            self.vacuum_logger.error('No matching explores found')
            raise Exception('No matching explores found')
        return info
